"""
API:

on_finish:
  data:
    column:
      - unique
      - no_nulls
      - values: [a, b, c]
"""


class TestSpec:
    """Convert dictionaries to a callable that runs assertions on products
    """
    def __init__(self, data, arg, client=None):
        pass

    def __call__(self):
        pass
